from enum import Enum
from typing import Dict, Any


class StopCriteria(Enum):
    EOS_TOKEN = "eos_token"
    FORCED_BY_HOOK = "forced_by_hook"
    TOKEN_LIMIT = "token_limit"


class EOSException(Exception):
    stop_criteria: StopCriteria
    extra: Dict[str, Any]

    def __init__(self, stop_criteria: StopCriteria, **kwargs: Any) -> None:
        self.stop_criteria = stop_criteria
        self.extra = kwargs

    def __getattr__(self, name: str) -> Any:
        if name in self.extra:
            return self.extra[name]
        raise AttributeError(f"{name} not found")

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, EOSException):
            return False

        return self.stop_criteria == value.stop_criteria and self.extra == self.extra



from enum import Enum
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import torch
from synthetic.exceptions import EOSException
from synthetic.commands import CommandExecution
from typing import Dict, Any, List


class HookType(Enum):
    ON_TOKEN = "on_token"
    ON_EOS = "on_eos"


@dataclass
class Event:
    step: int
    input_ids: torch.Tensor
    command_executions: List[CommandExecution] = field(default_factory=list)


class Hook(ABC):
    hook_type: HookType

    def __init__(self) -> None:
        pass

    @abstractmethod
    def __call__(self, event) -> Event:
        pass


## --- ON_TOKEN --------


@dataclass
class TokenEvent(Event):
    pass


class OnTokenHook(Hook):
    hook_type = HookType.ON_TOKEN

    @abstractmethod
    def __call__(self, event: TokenEvent) -> TokenEvent:
        pass


## --- ON_EOS ---------


@dataclass(kw_only=True)
class EOSEvent(Event):
    eos_exception: EOSException
    output: Dict[str, Any]


class OnEOSHook(Hook):
    hook_type = HookType.ON_EOS

    @abstractmethod
    def __call__(self, event: EOSEvent) -> EOSEvent:
        pass



from synthetic.hooks import TokenEvent, OnTokenHook
from synthetic.commands import (
    CommandSignature,
    Command,
    CommandInput,
    CommandOutput,
    CommandExecution,
)
from transformers import PreTrainedTokenizerBase
from typing import Dict
import torch


class CommandHook(OnTokenHook):

    command_signature: CommandSignature
    tokenizer: PreTrainedTokenizerBase
    commands: Dict[str, Command]
    device: str

    def __init__(
        self,
        command_signature: CommandSignature,
        tokenizer: PreTrainedTokenizerBase,
        commands: Dict[str, Command],
        device: str,
    ) -> None:
        super().__init__()
        self.command_signature = command_signature
        self.tokenizer = tokenizer
        self.commands = commands
        self.device = device

    def __call__(self, event: TokenEvent) -> TokenEvent:
        text = self.tokenizer.batch_decode(event.input_ids)[0]
        match = self.command_signature.match(text)

        if match:

            command = self.commands.get(match.command_name)

            if command is None:
                output = CommandOutput(
                    error="Invalid command name.", input_ids=event.input_ids
                )
            else:
                command_input = CommandInput(
                    input=match.input,
                    input_ids=event.input_ids,
                )

                output = command.execute(command_input)

                event.command_executions.append(
                    CommandExecution(
                        step=event.step,
                        command_name=match.command_name,
                        command_input=command_input,
                        command_output=output,
                    )
                )

            text = self.tokenizer.batch_decode(output.input_ids)[0]
            if output.error:
                text = self.command_signature.replace_partial_at_end(
                    text, f"Error: {output.error}"
                )
            elif output.output:
                text = self.command_signature.replace_partial_at_end(
                    text, output.output
                )
            else:
                raise ValueError(
                    "CommandOutput should contain at least one of 'error' or 'data' properties."
                )

            input_ids = (
                torch.tensor(self.tokenizer.encode(text), dtype=torch.int)
                .unsqueeze(0)
                .to(self.device)
            )

            event.input_ids = input_ids

        return event



from transformers import PreTrainedModel, PreTrainedTokenizerBase
from typing import List, Optional, Any, Dict, Tuple
from synthetic.hooks import (
    Hook,
    OnTokenHook,
    OnEOSHook,
    HookType,
    TokenEvent,
    EOSEvent,
    CommandHook,
)
from synthetic.exceptions import EOSException, StopCriteria
from synthetic.commands import Command, CommandSignature, CommandExecution
from jinja2 import Template, StrictUndefined, Undefined
import torch
from dataclasses import dataclass, field
from collections import Counter


@dataclass
class GenerationOutput:
    input: Dict[str, Any]
    step: int
    command_executions: List[CommandExecution]

    extra: Dict[str, Any] = field(default_factory=dict)

    def __init__(
        self,
        input: Dict[str, Any],
        step: int,
        command_executions: List[CommandExecution],
        **kwargs,
    ) -> None:
        self.input = input
        self.step = step
        self.command_executions = command_executions
        self.extra = kwargs

    def __getattr__(self, name: str) -> Any:
        if name in self.extra:
            return self.extra[name]
        raise AttributeError(f"{name} not found")


class Transformer:

    model: PreTrainedModel
    tokenizer: PreTrainedTokenizerBase
    device: str

    prompt_template: Template

    max_new_tokens: Optional[int]
    greedy: bool

    command_signature: CommandSignature

    _hooks: List[Hook]
    _commands: List[Command]

    def __init__(
        self,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        device: str = "cpu",
        eos_token: Optional[str] = None,
        prompt_template: Optional[str] = None,
        strict: bool = False,
        max_new_tokens: Optional[int] = None,
        greedy: bool = True,
        command_signature: CommandSignature = CommandSignature("<<", ":", "|", ">>"),
        hooks: List[Hook] = [],
        commands: List[Command] = [],
    ) -> None:

        self.model: PreTrainedModel = model.to(device)  # type: ignore
        self.tokenizer = tokenizer
        self.device = device

        if eos_token:
            self.eos_token = self.tokenizer.convert_tokens_to_ids(eos_token)  # type: ignore
        elif self.model.config.eos_token_id:
            self.eos_token = self.model.config.eos_token_id
        else:
            raise ValueError("No EOS token provided!")

        if prompt_template:
            undefined = StrictUndefined if strict else Undefined
            self.prompt_template = Template(prompt_template, undefined=undefined)
        else:
            self.prompt_template = Template("{{ inputs }}", undefined=StrictUndefined)

        self.max_new_tokens = max_new_tokens
        self.greedy = greedy

        self.command_signature = command_signature

        self._hooks = hooks
        self._commands = commands

    @staticmethod
    def _build_command_dict(commands: List[Command]) -> Dict[str, Command]:
        name_counts = Counter(cmd.name for cmd in commands)

        result: Dict[str, Command] = {}
        counters: Dict[str, int] = {}

        for cmd in commands:
            base_name = cmd.name
            if name_counts[base_name] == 1:
                # Unique name, no suffix
                key = base_name
            else:
                # Shared name, add suffix
                counters[base_name] = counters.get(base_name, 0) + 1
                key = f"{base_name}-{counters[base_name]}"
            result[key] = cmd

        return result

    @staticmethod
    def _sort_hooks(hooks: List[Hook]) -> Tuple[List[OnTokenHook], List[OnEOSHook]]:
        on_token_hooks = []
        on_eos_hooks = []

        for hook in hooks:
            if hook.hook_type == HookType.ON_TOKEN:
                on_token_hooks.append(hook)
            elif hook.hook_type == HookType.ON_EOS:
                on_eos_hooks.append(hook)
            else:
                raise ValueError("Failed to sort hooks as an invalid hook was found.")

        return (on_token_hooks, on_eos_hooks)

    @staticmethod
    def _run_on_token_hooks(event: TokenEvent, hooks: List[OnTokenHook]) -> TokenEvent:

        for hook in hooks:

            if not isinstance(hook, OnTokenHook):
                raise ValueError(f"Tried to run '{hook}' as an OnTokenHook.")

            event = hook(event)

        return event

    @staticmethod
    def _run_on_eos_hooks(event: EOSEvent, hooks: List[OnEOSHook]) -> EOSEvent:

        for hook in hooks:

            if not isinstance(hook, OnEOSHook):
                raise ValueError(f"Tried to run '{hook}' as an OnEOSHook.")

            event = hook(event)

        return event

    def generate(self, **kwargs: Any) -> GenerationOutput:

        command_dict = self._build_command_dict(self._commands)

        on_token_hooks, on_eos_hooks = self._sort_hooks(
            [
                CommandHook(
                    command_signature=self.command_signature,
                    tokenizer=self.tokenizer,
                    device=self.device,
                    commands=command_dict,
                )
            ]
            + self._hooks
        )

        command_executions: List[CommandExecution] = []

        prompt = self.prompt_template.render(**kwargs)

        input_ids = (
            torch.tensor(self.tokenizer.encode(prompt), dtype=torch.int)
            .unsqueeze(0)
            .to(self.device)
        )

        step = 0

        try:
            while self.max_new_tokens is None or step < self.max_new_tokens:

                step += 1

                outputs = self.model(
                    input_ids=input_ids,
                    use_cache=False,
                )

                logits = outputs.logits[:, -1, :]

                if self.greedy:
                    next_token = torch.argmax(logits, dim=-1)
                else:
                    probs = torch.softmax(logits, dim=-1)
                    next_token = torch.multinomial(probs, num_samples=1).squeeze(-1)

                if next_token.item() == self.eos_token:
                    raise EOSException(stop_criteria=StopCriteria.EOS_TOKEN)

                input_ids = torch.cat([input_ids, next_token.unsqueeze(-1)], dim=-1)

                token_event = TokenEvent(
                    input_ids=input_ids,
                    step=step,
                    command_executions=command_executions,
                )
                token_event = self._run_on_token_hooks(token_event, on_token_hooks)

                input_ids = token_event.input_ids
                command_executions = token_event.command_executions

            raise EOSException(stop_criteria=StopCriteria.TOKEN_LIMIT)

        except EOSException as e:

            eos_event = EOSEvent(
                step=step,
                eos_exception=e,
                input_ids=input_ids,
                command_executions=command_executions,
                output={"output": self.tokenizer.batch_decode(input_ids)[0]},
            )

            eos_event = self._run_on_eos_hooks(eos_event, on_eos_hooks)

        return GenerationOutput(
            input=kwargs,
            step=step,
            command_executions=eos_event.command_executions,
            **eos_event.output,
        )



from dataclasses import dataclass
from typing import Optional


@dataclass
class SignatureMatch:
    command_name: str
    input: str


class CommandSignature:
    start_string: str
    name_input_separation: str
    input_output_separation: str
    end_string: str

    def __init__(
        self,
        start_string: str,
        name_input_separation: str,
        input_output_separation: str,
        end_string: str,
    ) -> None:
        if not start_string:
            raise ValueError("The start_string should not be empty.")
        if not name_input_separation:
            raise ValueError("The name_input_separation should not be empty.")
        if not input_output_separation:
            raise ValueError("The input_output_separation should not be empty.")
        if not end_string:
            raise ValueError("The end_string should not be empty.")

        self.start_string = start_string
        self.name_input_separation = name_input_separation
        self.input_output_separation = input_output_separation
        self.end_string = end_string

    def match(self, text: str) -> Optional[SignatureMatch]:
        if not text.endswith(self.input_output_separation):
            return None

        # Extract the portion between start_string and the last input_output_separation
        before_sep = text[: -len(self.input_output_separation)]

        # Find the last start_string
        start_idx = before_sep.rfind(self.start_string)
        if start_idx == -1:
            return None

        inner = before_sep[start_idx + len(self.start_string) :]

        # Must contain name_input_separation
        if self.name_input_separation not in inner:
            return None

        name, input_str = inner.split(self.name_input_separation, 1)

        return SignatureMatch(command_name=name.strip(), input=input_str.strip())

    def dump_to_string(self, command_name: str, input: str, output: str) -> str:
        return (
            f"{self.start_string}"
            f"{command_name}{self.name_input_separation}{input}"
            f"{self.input_output_separation}{output}"
            f"{self.end_string}"
        )

    def replace_partial_at_end(self, text: str, output: str) -> str:
        m = self.match(text)
        if not m:
            return text

        # Find where the partial signature starts
        start_idx = text.rfind(self.start_string)
        # Replace it with the full signature
        return text[:start_idx] + self.dump_to_string(m.command_name, m.input, output)



from dataclasses import dataclass
from typing import Callable, Optional
import torch


@dataclass
class CommandInput:
    input: str
    input_ids: torch.Tensor


class CommandOutput:
    input_ids: torch.Tensor
    output: Optional[str]
    error: Optional[str]

    def __init__(
        self,
        input_ids: torch.Tensor,
        output: Optional[str] = None,
        error: Optional[str] = None,
    ) -> None:
        if output is None and error is None:
            raise ValueError(
                "At least one of the 'output' or 'error' arguments has to be not None."
            )

        self.input_ids = input_ids
        self.output = output
        self.error = error


@dataclass
class CommandExecution:
    step: int
    command_name: str
    command_input: CommandInput
    command_output: CommandOutput


class Command:
    name: str
    execute: Callable[[CommandInput], CommandOutput]

    def __init__(
        self, name: str, execute: Callable[[CommandInput], CommandOutput]
    ) -> None:
        self.name = name
        self.execute = execute
